""" 
Add a docstring
"""
import numpy as np
import pandas as pd
import casadi as cs
import matplotlib.pyplot as plt
from nloed import Model
from nloed import Design
####################################################################################################
# Block 1: Casadi Setup
####################################################################################################
#define input and parameter symbols
inputs = cs.SX.sym('inputs',2)
parameters = cs.SX.sym('parameters',8)

K_dr = cs.exp(parameters[0]) 
K_T = cs.exp(parameters[1])
B_o = cs.exp(parameters[2])
B = cs.exp(parameters[3])
K_m = cs.exp(parameters[4])
K_g = cs.exp(parameters[5])
Sigma_m = cs.exp(parameters[6])
Sigma_g = cs.exp(parameters[7])
m0 = 0.00002334410
g0 = 1468.31716701552

ccar = 1/( cs.sqrt(K_dr*(K_T/inputs[0]+1)**2+1) + cs.sqrt(K_dr)*(K_T/inputs[0]+1) )**2
c0 = (B_o + B*ccar/(K_m + ccar))
c1 = K_g*c0/Sigma_m
c2 = c1 - K_g*m0

gfp_mean = c1/Sigma_g +\
           c2/(Sigma_m - Sigma_g)*cs.exp(-Sigma_m*inputs[1]) +\
           (g0-c1/Sigma_g - c2/(Sigma_m - Sigma_g))*cs.exp(-Sigma_g*inputs[1])
#assume some hetroskedasticity, std_dev 5% of mean expression level
gfp_var = (0.05**2)*gfp_mean**2
#link the deterministic model to the sampling statistics (here normal mean and variance)
gfp_stats = cs.vertcat(gfp_mean, gfp_var)
#create a casadi function mapping input and parameters to sampling statistics (mean and var)
gfp_model = cs.Function('GFP',[inputs,parameters],[gfp_stats])

K_dr_val = 0.00166160831
K_T_val = 11979.25516474195
B_o_val = 0.00137146821
B_val = 0.03525605965
K_m_val = 0.28169444728
K_g_val = 68.61725640178
Sigma_m_val = 0.01180105629
Sigma_g_val = 0.01381497437

fit_params = np.log([K_dr_val, K_T_val, B_o_val, B_val, K_m_val, K_g_val, Sigma_m_val, Sigma_g_val])

# enter the function as a tuple with label indicating normal error, into observation list
observ_list = [(gfp_model,'Normal')]
#create names for inputs
input_names = ['Light','Time']
#create names for parameters
parameter_names = ['K_dr', 'K_T', 'B_o', 'B', 'K_m', 'K_g', 'Sigma_m', 'Sigma_g']
#instantiate nloed model class
model_object = Model(observ_list,input_names,parameter_names)

#set up the design algorithm to use continuous (continuous) optimization with two unique inputs points
continuous_inputs={'Inputs':['Light','Time'],
                   'Bounds':[(.01,4095),(0,6*60)],
                   'Structure':[['L1','T1'],
                                ['L2','T2'],
                                ['L3','T3'],
                                ['L4','T4'],
                                ['L5','T5'],
                                ['L6','T6'],
                                ['L7','T7'],
                                ['L8','T8'],
                                ['L9','T9'],
                                ['L10','T10']],
                    'Initial':[[0.1,350],
                                [10,350],
                                [100,350],
                                [1000,350],
                                [4090,350],
                                [4090,1],
                                [4090,10],
                                [4090,100],
                                [4090,300],
                                [4090,350]]}

#set up fixed design dictionary
#fixed_dict ={'Weight':0.5,'Design':init_design}
# generate the optimal discreteimate (relaxed) design
design_object = Design(model_object,fit_params,'D',
                       continuous_inputs = continuous_inputs)
#extract the relaxed design structure
relaxed_design = design_object.relaxed()
#***hidden_start****
print('')
print(relaxed_design)
print('')
#***hidden_end****
#set the sample size to 30
sample_size = 15
#generate a rounded exact design 
exact_design = design_object.round(sample_size)
#***hidden_start****
print('')
print(exact_design)
print('')
#***hidden_end****

####################################################################################################
# Block 7: fit combined data
####################################################################################################
#***hidden_start****
#generate some data for optimal design and combine 
intu_design = pd.DataFrame({'Light':[.1]+[63]+[254]+[1022]+[4095],
                          'Variable':['GFP']*5,
                          'Replicats':[3]*5}) 
intu_data = model_object.sample(intu_design,true_param)
optimal_data = model_object.sample(exact_design,true_param)
combined_data = pd.concat([init_data, optimal_data], ignore_index=True)
total_null_data = pd.concat([init_data, intu_data], ignore_index=True)
#***hidden_end****
total_null_design = pd.concat([init_design, intu_design], ignore_index=True)
#combine the initial and optimal design
combined_design = pd.concat([init_design, exact_design], ignore_index=True)

eval_options={'Method':'Asymptotic',
              'Covariance':True}
# call evaluate to compute the asymptotic covariance
asymptotic_covariance = model_object.evaluate(combined_design,true_param,eval_options)
#compute the asymptotic upper and lower 95% bounds
asymptotic_lower_bound = fit_params - 2*np.sqrt(np.diag(asymptotic_covariance))
asymptotic_upper_bound = fit_params + 2*np.sqrt(np.diag(asymptotic_covariance))
#***hidden_start****
fim_bnds_array = np.vstack((asymptotic_lower_bound,
                            fit_params,
                            asymptotic_upper_bound)).T
fim_bnds = pd.DataFrame(np.exp(fim_bnds_array),
                        index=['Alpha0','Alpha','n','K'],
                        columns=['Lower','Estimate','Upper'])
print('')
print(fim_bnds)
print('')

eval_options={'Method':'Asymptotic',
              'Covariance':True}
# call evaluate to compute the asymptotic covariance
asymptotic_covariance = model_object.evaluate(total_null_design,true_param,eval_options)
#compute the asymptotic upper and lower 95% bounds
asymptotic_lower_bound = fit_params - 2*np.sqrt(np.diag(asymptotic_covariance))
asymptotic_upper_bound = fit_params + 2*np.sqrt(np.diag(asymptotic_covariance))
#***hidden_start****
fim_bnds_array = np.vstack((asymptotic_lower_bound,
                            fit_params,
                            asymptotic_upper_bound)).T
fim_bnds = pd.DataFrame(np.exp(fim_bnds_array),
                        index=['Alpha0','Alpha','n','K'],
                        columns=['Lower','Estimate','Upper'])
print('')
print(fim_bnds)
print('')




#request contours and use a simple initial search
fit_options={'Confidence':'Contours'}
#fit the model to the initial data
fit_opt = model_object.fit(combined_data,start_param = fit_params, options=fit_options)
#extract the parameter values
fit_params = fit_info['Estimate'].to_numpy().flatten()
#***hidden_start****
print('')
print(np.exp(fit_params))
print('')
#***hidden_end****

####################################################################################################
# Block 8: CI comparison for initial data
####################################################################################################
#***hidden_start****
lik_bnds_array = np.hstack((fit_info['Lower'].to_numpy().T,
                        fit_info['Estimate'].to_numpy().T,
                        fit_info['Upper'].to_numpy().T))
lik_bnds = pd.DataFrame(np.exp(lik_bnds_array),
                        index=['Alpha0','Alpha','n','K'],
                        columns=['Lower','Estimate','Upper'])
print('')
print(lik_bnds)
print('')
#***hidden_end****
#request the asymptotic covariance matrix
eval_options={'Method':'Asymptotic',
              'Covariance':True}
# call evaluate to compute the asymptotic covariance
asymptotic_covariance = model_object.evaluate(combined_design,fit_params,eval_options)
#compute the asymptotic upper and lower 95% bounds
asymptotic_lower_bound = fit_params - 2*np.sqrt(np.diag(asymptotic_covariance))
asymptotic_upper_bound = fit_params + 2*np.sqrt(np.diag(asymptotic_covariance))
#***hidden_start****
fim_bnds_array = np.vstack((asymptotic_lower_bound,
                            fit_params,
                            asymptotic_upper_bound)).T
fim_bnds = pd.DataFrame(np.exp(fim_bnds_array),
                        index=['Alpha0','Alpha','n','K'],
                        columns=['Lower','Estimate','Upper'])
print('')
print(fim_bnds)
print('')
# #***hidden_end****
# #generate 500 simulated datasets
# mc_data = model_object.sample(init_design,fit_params,design_replicats=500)
# #fit the 500 datasets
# mc_fits = model_object.fit(mc_data, start_param = fit_params)
# #compute the Monte Carlo covariance estimate
# mc_covariance = np.cov(np.exp(mc_fits['Estimate'].to_numpy().T))
# #compute the asymptotic upper and lower 95% bounds
# mc_lower_bound = fit_params - 2*np.sqrt(np.diag(mc_covariance))
# mc_upper_bound = fit_params + 2*np.sqrt(np.diag(mc_covariance))
# #***hidden_start****
# mc_bnds_array = np.vstack((mc_lower_bound,
#                            fit_params,
#                            mc_upper_bound)).T
# mc_bnds = pd.DataFrame(np.exp(mc_bnds_array),
#                         index=['Alpha0','Alpha','n','K'],
#                         columns=['Lower','Estimate','Upper'])
# print('')
# print(mc_bnds)
# print('')
# #***hidden_end****

####################################################################################################
# Block 5: plot model predictions and initial data
####################################################################################################
#convert the covariance matrix to numpy
covariance_matrix = asymptotic_covariance.to_numpy()
#generate predictions with error bars fdor a random selection of inputs)
prediction_inputs = pd.DataFrame({'Light':np.linspace(0.1,10,100),
                                  'Variable':['GFP']*100})
#request prediction and observation intervals
prediction_options = {'PredictionInterval':True,
                      'ObservationInterval':True}
#generate predictions and intervals
predictions = model_object.predict(prediction_inputs,
                                   fit_params,
                                   covariance_matrix = covariance_matrix,
                                   options=prediction_options)
#create plot
fig, ax = plt.subplots()
#plot observation interval
ax.fill_between(predictions['Inputs','Light'],
                predictions['Observation','Lower'],
                predictions['Observation','Upper'],
                alpha=0.3,
                color='C1')
#plot prediction interval
ax.fill_between(predictions['Inputs','Light'],
                predictions['Prediction','Lower'],
                predictions['Prediction','Upper'],
                alpha=0.4,
                color='C0')
#plot mean model prediction
ax.plot(predictions['Inputs','Light'], predictions['Prediction','Mean'], '-')
#plot initial dataset
ax.plot(combined_data['Light'], combined_data['Observation'], 'o', color='C1')
ax.set_xlabel('Light')
ax.set_ylabel('GFP')
plt.show()

perc0=(fim_bnds0['Upper']-fim_bnds0['Lower'])/fim_bnds0['Estimate']
perc=(fim_bnds['Upper']-fim_bnds['Lower'])/fim_bnds['Estimate']

labels = ['Alpha0','Alpha','n','K']
x = np.arange(len(labels))  # the label locations
width = 0.35  # the width of the bars
fig, ax = plt.subplots()
rects1 = ax.bar(x - width/2, 100*perc0.to_numpy(), width, label='Initial')
rects2 = ax.bar(x + width/2, 100*perc.to_numpy(), width, label='Optimal')
ax.set_ylabel('% Error')
ax.set_title('Diameter of 95% CI Expressed in % Error')
ax.set_xticks(x)
ax.set_xticklabels(labels)
ax.legend()
fig.tight_layout()
plt.show()


#!/usr/bin/env python
# coding: utf-8

# # Sampling and Fitting the Hill Function Model
# 

# In[24]:


#%%capture
from N2_optimizing_a_design import *


# ### Sampling the Model using Simulated Data
# 
# 

# In[25]:


data = hill_model.sample(exact_design,nominal_param)


# 

# In[26]:


print(data)


# ### Fitting the Model with Maximum Likelihood

# In[34]:


fit_options={'Confidence':'Profiles',  
             'InitParamBounds':[(-1,4),(-1,4),(-1,4),(-1,4)],
             'InitSearchNumber':7,
             'SearchBound':5.}


# 

# In[35]:


fit_info = hill_model.fit(data, options=fit_options)
print(fit_info['Estimate'])


# In[33]:


print(fit_info['Lower'])
print(fit_info['Upper'])


# 

# In[29]:


fit_params = fit_info['Estimate'].to_numpy().flatten()
print(np.exp(fit_params))


# In[30]:


print(np.exp(nominal_param))


# In[ ]:







import pytest
import nloed as nl 
import casadi as cs 

@pytest.fixture
def linear_model():
    #define input and parameters
    x = cs.SX.sym('x',2)
    xnames = ['x1','x2']
    p = cs.SX.sym('p',2)
    pnames = ['p1','p2']
    #define predictor symbol
    predictor = 1 + p[0]*x[0] + p[1]*x[1]
    #define stats
    stats = cs.vertcat(predictor, 0.1*predictor**2)
    #create a casadi function
    y = cs.Function('y', [x,p], [stats])
    #observ struct
    observ = [(y,'Normal')]
    #instantiate model class
    linear_model = nl.Model(observ,xnames,pnames)

    return linear_model


@pytest.fixture(params=["Bounds","Grid", "Candidates"])
def discrete_input_struct(request):
    if request.param == "Bounds":
        input_struct={'Inputs':['x1','x2'],
                      'Bounds':[[-1,1],[-1,1]]}
    elif request.param == "Grid":
        input_struct={'Inputs':['x1','x2'],
                      'Grid':[[-1,-1],[1,-1],[-1,1],[1,1]]}
    elif request.param == "Candidates":
        input_struct={'Inputs':['x1','x2'],
                      'Candidates':[[-1,0,1],[-1,0,1]]}

    return input_struct

@pytest.fixture
def discrete_design(linear_model,discrete_input_struct):
    #define input struct
    nominal_param = [1,1]
    objective = 'D'

    # generate the optimal approximate (relaxed) design
    linear_design = nl.Design(linear_model,
                              nominal_param,
                              objective,
                              discrete_inputs=discrete_input_struct)

    return linear_design


""" 
Add a docstring
"""
import numpy as np
import pandas as pd
import casadi as cs
import matplotlib.pyplot as plt
from nloed import Model
from nloed import Design
####################################################################################################
# Block 1: Casadi Setup
####################################################################################################
#define input and parameter symbols
inputs = cs.SX.sym('inputs')
parameters = cs.SX.sym('parameters',5)

K_dr = cs.exp(parameters[0]) 
K_T = cs.exp(parameters[1])
B_o_prime = cs.exp(parameters[2])
B_prime = cs.exp(parameters[3])
K_m = cs.exp(parameters[4])

ccar = 1/( cs.sqrt(K_dr*(K_T/inputs+1)**2+1) + cs.sqrt(K_dr)*(K_T/inputs+1) )**2
gfp_mean = (B_o_prime + B_prime*ccar/(K_m + ccar))

#assume some hetroskedasticity, std_dev 5% of mean expression level
gfp_var = (0.05**2)*gfp_mean**2
#link the deterministic model to the sampling statistics (here normal mean and variance)
gfp_stats = cs.vertcat(gfp_mean, gfp_var)
#create a casadi function mapping input and parameters to sampling statistics (mean and var)
gfp_model = cs.Function('GFP',[inputs,parameters],[gfp_stats])

# K_g_val = 68.61725640178
# Sigma_m_val = 0.01180105629
# Sigma_g_val = 0.01381497437
# K_dr_val = 0.00166160831
# K_T_val = 11979.25516474195
# K_m_val = 0.28169444728
# B_o_val = 0.00137146821
# B_val = 0.03525605965

K_g_val = 1.6260543
Sigma_m_val = 0.00490364
Sigma_g_val = 0.046474403
K_dr_val = 4.04461938
K_T_val = 11306.84788
K_m_val = 0.000149759
B_o_val = 0.073858518
B_val = 1.8532546

B_o_prime_val = B_o_val*K_g_val/Sigma_m_val/Sigma_g_val
B_prime_val = B_val*K_g_val/Sigma_m_val/Sigma_g_val

fit_params = np.log([K_dr_val, K_T_val, B_o_prime_val, B_prime_val, K_m_val])

# enter the function as a tuple with label indicating normal error, into observation list
observ_list = [(gfp_model,'Normal')]
#create names for inputs
input_names = ['Light']
#create names for parameters
parameter_names = ['K_dr', 'K_T', 'B_o_prime', 'B_prime', 'K_m']
#instantiate nloed model class
static_model = Model(observ_list,input_names,parameter_names)

# #generate predictions with error bars fdor a random selection of inputs)
# prediction_inputs = pd.DataFrame({'Light':np.linspace(0.1,4095,100),
#                                   'Variable':['GFP']*100})
# #generate predictions and intervals
# predictions = static_model.predict(prediction_inputs,
#                                    fit_params,
#                                    options ={'Sensitivity':True})
# #create plot
# fig, ax = plt.subplots()
# #plot mean model prediction
# ax.plot(predictions['Inputs','Light'], predictions['Prediction','Mean'], '-')
# ax.set_xlabel('Light')
# ax.set_ylabel('GFP')
# plt.show()

obs1 = [1101,
        1246,
        883,
        2571,
        1768,
        1219,
        577,
        8029,
        11104,
        9792,
        16450,
        24290,
        39988,
        6480,
        33019]
lght_1 = [.1]*3+[63]*3+[254]*3+[1022]*3+[4095]*3

obs2 = [837,
        733,
        2807,
        2816,
        1857,
        4455,
        4408,
        6779,
        3325,
        4343,
        3095,
        7030,
        10583,
        11957,
        17313]
lght_2 = [.1]*2+[256]*4+[597]*5+[4095]*4

obs3 = [975,
        827,
        1103,
        1833,
        1743,
        1128,
        2459,
        3889,
        7126,
        2610,
        6837,
        10089,
        9367,
        9838]
lght_3 = [110]*3+[192]*3+[335]*2+[585]*3+[2046]*3

init_data = pd.DataFrame({'Light':lght_1 + lght_2 + lght_3,
                          'Variable':['GFP']*(15*2+14),
                          'Observation':obs1 + obs2 + obs3}) 
            
init_design = pd.DataFrame({'Light':lght_1 + lght_2 + lght_3,
                            'Variable':['GFP']*(15*2+14),
                            'Replicates':[1]*(15*2+14)}) 


#set up the design algorithm to use continuous (continuous) optimization with two unique inputs points
continuous_inputs={'Inputs':['Light'],
                   'Bounds':[(.01,4095)],
                   'Structure':[['x1'],
                                ['x2'],
                                ['x3'],
                                ['x4'],
                                ['x5']],
                    'Initial':[[0.1],
                               [100],
                               [300],
                               [1000],
                               [4000]]}
#set up fixed design dictionary
fixed_dict ={'Weight':.75,'Design':init_design}
# generate the optimal discreteimate (relaxed) design
design_object = Design(static_model,fit_params,'D',
                       fixed_design = fixed_dict,
                       continuous_inputs = continuous_inputs)
#extract the relaxed design structure
relaxed_design = design_object.relaxed()
#***hidden_start****
print('')
print(relaxed_design)
print('')
#***hidden_end****
#set the sample size to 30
sample_size = 16
#generate a rounded exact design 
exact_design = design_object.round(sample_size)
#***hidden_start****
print('')
print(exact_design)
print('')
#***hidden_end****


eval_options={'Method':'Asymptotic',
              'Covariance':True,
              'FIM':True}
# call evaluate to compute the asymptotic covariance
diagnostics0 = static_model.evaluate(init_design,fit_params,eval_options)
print(np.linalg.cond(diagnostics0['FIM']))
asymptotic_covariance = diagnostics0['Covariance']
#compute the asymptotic upper and lower 95% bounds
asymptotic_lower_bound = fit_params - 2*np.sqrt(np.diag(asymptotic_covariance))
asymptotic_upper_bound = fit_params + 2*np.sqrt(np.diag(asymptotic_covariance))
#***hidden_start****
fim_bnds_array = np.vstack((asymptotic_lower_bound,
                            fit_params,
                            asymptotic_upper_bound)).T
fim_bnds0 = pd.DataFrame(np.exp(fim_bnds_array),
                        index=parameter_names,
                        columns=['Lower','Estimate','Upper'])
print('')
print(fim_bnds0)
print('')

combined_design = pd.concat([init_design, exact_design], ignore_index=True)
# call evaluate to compute the asymptotic covariance
diagnostics_opt = static_model.evaluate(combined_design,fit_params,eval_options)
print(np.linalg.cond(diagnostics_opt['FIM']))
asymptotic_covariance = diagnostics_opt['Covariance']
#compute the asymptotic upper and lower 95% bounds
asymptotic_lower_bound = fit_params - 2*np.sqrt(np.diag(asymptotic_covariance))
asymptotic_upper_bound = fit_params + 2*np.sqrt(np.diag(asymptotic_covariance))
#***hidden_start****
fim_bnds_array = np.vstack((asymptotic_lower_bound,
                            fit_params,
                            asymptotic_upper_bound)).T
fim_bnds_opt = pd.DataFrame(np.exp(fim_bnds_array),
                        index=parameter_names,
                        columns=['Lower','Estimate','Upper'])
print('')
print(fim_bnds_opt)
print('')

####################################################################################################
# Block 2: Casadi Setup
####################################################################################################

lght_1 = [.1]*2+[256]*4+[597]*5+[4095]*4
lght_2 = [110]*3+[192]*3+[335]*2+[585]*3+[2046]*3
lght_3 = [.1]*3+[63]*3+[254]*3+[1022]*3+[4095]*3

tmpts_1 = [1*60]*3 + [2*60]*3 + [3*60]*3 + [4*60]*3 + [5*60]*3

init_design = pd.DataFrame({'Time':[6*60]*(15*2+14)+[0.5*60]*15+[1.8*60]*15+tmpts_1*2,
                            'Light1':lght_1 + lght_2 + lght_3*3 + [4095]*15 +[2046]*15,
                            'Light2':lght_1 + lght_2 + lght_3*3 + [4095]*15 +[2046]*15,
                            'Light3':lght_1 + lght_2 + lght_3*3 + [4095]*15 +[2046]*15,
                            'Variable':['GFP_1']*(15*2+14+15*2+15*2),
                            'Replicates':[1]*(15*6+14)}) 

#define input and parameter symbols
inputs = cs.SX.sym('inputs',4)
parameters = cs.SX.sym('parameters',3)

K_g = cs.exp(parameters[0])
Sigma_m = cs.exp(parameters[1])
Sigma_g = cs.exp(parameters[2])
m0 = np.mean([19.17,11.82,0.00013])
g0 = np.mean([2713,3560,2436])


# B_o_prime_val = B_o_val*K_g_val/Sigma_m_val/Sigma_g_val
# B_prime_val = B_val*K_g_val/Sigma_m_val/Sigma_g_val

observ_list=[]

indx = [0]*5 + [1]*4 + [2]*7

for i in range(16):
    ccar = 1/( cs.sqrt(K_dr_val*(K_T_val/inputs[indx[i]+1]+1)**2+1) + cs.sqrt(K_dr_val)*(K_T_val/inputs[indx[i]+1]+1) )**2
    c0 = (B_o_val + B_val*ccar/(K_m_val + ccar))
    c1 = K_g*c0/Sigma_m
    c2 = c1 - K_g*m0

    gfp_mean = c1/Sigma_g +\
            c2/(Sigma_m - Sigma_g)*cs.exp(-Sigma_m*inputs[0]) +\
            (g0-c1/Sigma_g - c2/(Sigma_m - Sigma_g))*cs.exp(-Sigma_g*inputs[0])
    #assume some hetroskedasticity, std_dev 5% of mean expression level
    gfp_var = (0.05**2)*gfp_mean**2
    #link the deterministic model to the sampling statistics (here normal mean and variance)
    gfp_stats = cs.vertcat(gfp_mean, gfp_var)
    #create a casadi function mapping input and parameters to sampling statistics (mean and var)
    gfp_model = cs.Function('GFP_'+str(i+1),[inputs,parameters],[gfp_stats])
    # enter the function as a tuple with label indicating normal error, into observation list
    observ_list.append((gfp_model,'Normal'))

fit_params = np.log([K_g_val, Sigma_m_val, Sigma_g_val])

#create names for inputs
input_names = ['Time','Light1','Light2','Light3']
#create names for parameters
parameter_names = ['K_g', 'Sigma_m', 'Sigma_g']
#instantiate nloed model class
dynamic_model = Model(observ_list,input_names,parameter_names)

discrete_inputs={'Inputs':['Light1','Light2','Light3'],
                 'Grid':[[0.01,1942,4095]]}

#set up the design algorithm to use continuous (continuous) optimization with two unique inputs points
continuous_inputs={'Inputs':['Time'],
                   'Bounds':[(0,6*60)],
                   'Structure':[['T1'],
                                ['T2'],
                                ['T3']],
                   'Initial':[[1],
                              [3*60],
                              [6*60]]}

#set up fixed design dictionary
fixed_dict ={'Weight':0.875,'Design':init_design}
# generate the optimal discreteimate (relaxed) design
opts= {'LockWeights':False}
observ_groups = [['GFP_'+str(i+1) for i in range(16)]]
design_object = Design(dynamic_model,fit_params,'D',
                       discrete_inputs = discrete_inputs,
                       continuous_inputs = continuous_inputs,
                       fixed_design = fixed_dict,
                       observ_groups = observ_groups,
                       options=opts)
#extract the relaxed design structure
relaxed_design = design_object.relaxed()
#***hidden_start****
print('')
print(relaxed_design)
print('')
#***hidden_end****
#set the sample size to 30
sample_size = 15
#generate a rounded exact design 
exact_design = design_object.round(sample_size)
#***hidden_start****
print('')
print(exact_design)
print('')
#***hidden_end****

eval_options={'Method':'Asymptotic',
              'Covariance':True,
              'FIM':True}
# call evaluate to compute the asymptotic covariance
diagnostics0 = dynamic_model.evaluate(init_design,fit_params,eval_options)
print(np.linalg.cond(diagnostics0['FIM']))
asymptotic_covariance = diagnostics0['Covariance']
#compute the asymptotic upper and lower 95% bounds
asymptotic_lower_bound = fit_params - 2*np.sqrt(np.diag(asymptotic_covariance))
asymptotic_upper_bound = fit_params + 2*np.sqrt(np.diag(asymptotic_covariance))
#***hidden_start****
fim_bnds_array = np.vstack((asymptotic_lower_bound,
                            fit_params,
                            asymptotic_upper_bound)).T
fim_bnds0 = pd.DataFrame(np.exp(fim_bnds_array),
                        index=parameter_names,
                        columns=['Lower','Estimate','Upper'])
print('')
print(fim_bnds0)
print('')

combined_design = pd.concat([init_design, exact_design], ignore_index=True)
# call evaluate to compute the asymptotic covariance
diagnostics_opt = dynamic_model.evaluate(combined_design,fit_params,eval_options)
print(np.linalg.cond(diagnostics_opt['FIM']))
asymptotic_covariance = diagnostics_opt['Covariance']
#compute the asymptotic upper and lower 95% bounds
asymptotic_lower_bound = fit_params - 2*np.sqrt(np.diag(asymptotic_covariance))
asymptotic_upper_bound = fit_params + 2*np.sqrt(np.diag(asymptotic_covariance))
#***hidden_start****
fim_bnds_array = np.vstack((asymptotic_lower_bound,
                            fit_params,
                            asymptotic_upper_bound)).T
fim_bnds_opt = pd.DataFrame(np.exp(fim_bnds_array),
                        index=parameter_names,
                        columns=['Lower','Estimate','Upper'])
print('')
print(fim_bnds_opt)
print('')


#!/usr/bin/env python
# coding: utf-8

# # Optimal Design for the 2-Input Linear Regression Model

# This example demonstrates how to generate an optimal design for the 2-input linear regression model dicussed in the previous notebook. To start we, import the *N1_model_creation* notebook so that the NLoed Model instance is available here.

# In[1]:


from scripts.N1_model_creation import *


# ###  Instantiating a Design Object
# 
# To create the NLoed design object we need to specify some aspects of the design we wish to generate, related to the physical constraints and the numerical formulation that will be used. We start by specifying the nominal parameter values and the objective function we will use in this design. Here we set the nominal parameters to all be $1$ and we use a D-optimal objective:

# In[ ]:


nominal_param = [1,1,1]
objective = 'D'


# We now need to specify how the Design algorithm will treat the inputs to the model. The main choice here is whether the model inputs are handled discretly or continuously. In this example we choose a continuous approach, see the NLoed manual and background documentation for further discussion and comparison. To have the Design object handle both inputs, $x1$ and $x2$ continuously we need to creat a continuous inputs dictionary that specifies the bounds of both inputs and the number of unique levels of each input the optimization algorithm will consider. Here we create a continuous_inputs dictionary which bounds both inputs between $-1$ and $1$ and allows for three unique pairs of input vectors to be considered in the design optimization (inidcated by the '*_lv#'* suffix in the *Structure* field):

# In[ ]:


continuous_inputs={'Inputs':['Input1','Input2'],
                   'Bounds':[(-1,1),(-1,1)],
                   'Structure':[['Input1_lv1','Input2_lv1'],
                                ['Input1_lv2','Input2_lv2'],
                                ['Input1_lv3','Input2_lv3']]}


# We now pass all of the design specification along with the model object, *linear_model*, into the NLoed Design constructor in order to instantiate a design object:

# In[ ]:


optimal_design = nl.Design(linear_model, nominal_param, objective, continuous_inputs=continuous_inputs)


# ### Viewing the Resulting Design
# 
# Once the Design object is created, the object instnace contains in it an optimal relaxed design which serves as an archetype for the given model and design scenario. Relaxed design's do not specify a specific sample size as they use real-valued weights to indicated the number of observations to be taken in a given set of conditions. However, somtimes it is useful to view the relaxed design and its corresponding weights as follows using the *.relaxed()* function:

# In[ ]:


optimal_design.relaxed()


# In order to generate an implementable design, the relaxed design needs to be rounded to an exact design. This can be done using the Desin class's *.round()* function as follows:

# In[ ]:


sample_size = 9
optimal_design.round(sample_size)


# The returned objects for the both the *.relaxed()* and *.round()* functions are dataframes making them easy to export to Excel or text files. The dataframes containing designs can also be used as input to user-callable functions of NLoed's Model class for data simulation and evauation of design performance on various metrics. See the other notebook examples for more details.



import pytest

def test_design_instantiation(discrete_design):

    assert discrete_design.num_models == 1
    assert discrete_design.input_dim == 2
    assert discrete_design.observ_dim == 1
    
    





## Listing 1 ##############################################
import nloed as nl, casadi as cs
import numpy as np, pandas as pd
#CasADi input and parameter symbols
x = cs.SX.sym('x',1)
theta = cs.SX.sym('theta',4)
#Transform; ensures parameter positivity
q=cs.exp(theta)
#Expressions for the sampling statistics
mean = (q[0] + q[1]*x[0]**q[2]/
              (q[3]**q[2] + x[0]**q[2]))
s = 0.059
var = (s*mean)**2
#Create a sampling statistics vector
stats = cs.vertcat(mean, var)
#Create a CasADi function for f(.)
func = cs.Function('GFP',[x,theta],[stats])

## Listing 2 ##############################################
#Input and parameter name lists
xname = ['Light']
tname = ['ln_A0','ln_A','ln_N','ln_K']
#Assign a normal dist. to the output
obs_var = [(func,'Normal')]
#Call the NLoed Model constructor
model = nl.Model(obs_var,xname,tname)

## Listing 3 ##############################################
#Initial green light levels (%)
levels0 = [0.01, 0.01, 0.01,
           1.5,  1.5,  1.5,
           6,    6,    6,
           25,   25,   25,
           100,  100,  100]
#Initial GFP observations (a.u.)
obs0 = [ 580,  544,   543,
         766,  699,   570,
         2064, 2171,  4598,
         8814, 9919,  8860,
         9584, 10295, 10672]
#Dataframe containing initial dataset
data0 = pd.DataFrame(
           {'Light':levels0,
            'Variable':['GFP']*15,
            'Observation':obs0}) 
#Set fit options
opts={'InitParamBounds':[(3,8),(5,10),
                         (-1,1),(1,6)],
      'InitSearchNumber':7,
      'SearchBound':5.}
#Fit the model to get an estimate
est = model.fit(data0, options=opts)
#Extract parameter estimate vector
param = est.to_numpy().flatten()

## Listing 4 ##############################################
#Specify input dictionary for the design
inp = {'Inputs':['Light'],
      'Bounds':[(.01,100.)],
      'Structure':[['L1'],
                    ['L2'],
                    ['L3'],
                    ['L4']],
      'Initial':[[0.01],
                  [5.],
                  [20.],
                  [100.]]}
#Declare init design dataframe
design0 = pd.DataFrame({
            'Light':[.01,1.5,6,25,100],
            'Variable':['GFP']*5,
            'Replicates':[3]*5}) 
#Set the fixed design information
init = {'Weight':0.5,'Design':design0}
#Set objective type
obj = 'D'
#Instantiate the design object
opt_des = nl.Design(model,param,obj,
                  fixed_design=init,
                  continuous_inputs=inp)
#Generate a rounded exact design 
design1 = opt_des.round(15)

## Listing 5 ##############################################
#Merge initial and optimal designs
total = pd.concat([design0, design1],
                    ignore_index=True)
#Compute the design metrics
metrics = model.evaluate(total, param)
#Extract the covariance matrix
cov = metrics['Covariance'].to_numpy()
#Set the list of input vectors
inputs = pd.DataFrame({
        'Light':np.linspace(.1,100,100),
        'Variable':['GFP']*100})
#set options for prediction method
opts = {'Method':'Delta',
        'ObservationInterval':True}
#Generate predictions and intervals
pred = model.predict(inputs, param,
                covariance_matrix = cov,
                options=opts)

## Supplemental Listing 1 ##############################################
#optimal green light levels
levels1 = \
[.001]*4+[2.88]*5+[8.33]*4+[100]*2
#optimal GFP observations (a.u.)
obs1 = [433,  477,  441,  604,   1032,
        823,  849,  792,  954,   3555,
        2039, 3384, 3740, 10321, 11534]
#dataframe containing initial dataset
data1 = pd.DataFrame(
            {'Light':levels0+levels1,
            'Variable':['GFP']*30,
            'Observation':obs0+obs1}) 
#fit the model
est = model.fit(data1, start_param=param)
#extract parameter estimate vector
param = est['Estimate'].to_numpy().flatten()

## Supplemental Listing 2 ##############################################
import matplotlib.pyplot as plt
import scipy.stats as st
#compute  # of std. deviations for a 95% interval
thresh = st.norm.ppf(0.95)
#compute the covariance for the initial design
covariance_init = model.evaluate(design0, param)
#compute the standard parameter error for initial design
stderr_init = np.sqrt(np.diag(covariance_init['Covariance']))
#compute upper/lower confidence bounds for initial design
upper_bnd_init = param + thresh*stderr_init
lower_bnd_init = param - thresh*stderr_init
#combine the initial and optimal designs
opt_design_tot = pd.concat([design0, design1], ignore_index=True)
#compute the covariance for the combined design
covariance_opt = model.evaluate(opt_design_tot, param)
#compute the parameter std. error for the combined design
stderr_opt = np.sqrt(np.diag(covariance_opt['Covariance']))
#compute the upper/lower confidence bounds for combined design
upper_bnd_opt = param + thresh*stderr_opt
lower_bnd_opt = param - thresh*stderr_opt
#replicate the initial design for the duplicated scenario
rep_init_design_tot = pd.concat([design0, design0], ignore_index=True)
#compute the covariance for the replicated initial design
covariance_rept = model.evaluate(rep_init_design_tot, param)
#compute the parameter std. error for replicated initial design
stderr_rept = np.sqrt(np.diag(covariance_rept['Covariance']))
#compute the upper/lower confidence bounds for replicated design
upper_bnd_rept = param + thresh*stderr_rept
lower_bnd_rept = param - thresh*stderr_rept

## Supplemental Listing 3 ##############################################
#exponentiate the final parameter estimates to convert
#from log scale 
exp_vals = np.exp(param)
#exponentiate the confidence bounds for each scenario and compute
#the normalized size of the intervals in the original scale
exp_diff_opt = (np.exp(upper_bnd_opt)-np.exp(lower_bnd_opt))/exp_vals
exp_diff_init = (np.exp(upper_bnd_init)-np.exp(lower_bnd_init))/exp_vals
exp_diff_rept = (np.exp(upper_bnd_rept)-np.exp(lower_bnd_rept))/exp_vals
#set the width of the bars
width = 0.25 
#create plot x labels names
labels = [r'$\alpha_o$',r'$\alpha$',r'n', r'K']
#set the x axis settings
x = np.arange(len(labels)) 
# create the figure
fig, ax = plt.subplots()
#plot the initial design CI interval sizes
rects1 = ax.bar(x - width-0.01, exp_diff_init.flatten(), width, label='Initial Design')
#plot the duplicated initial design CI interval sizes
rects2 = ax.bar(x, exp_diff_rept.flatten(), width, label='Repeated Initial Design')
#plot the combined (optimal + initial) design CI interval sizes
rects3 = ax.bar(x + width+0.01, exp_diff_opt.flatten(), width, label='Optimal Design')
#set y axis title
ax.set_ylabel('Interval Size as \% of MLE Parameter Value')
ax.set_xticks(x)
ax.set_xticklabels(labels)
ax.legend(loc=2)
fig.tight_layout()
#generate plot
plt.show()

## Supplemental Listing 4 ##############################################
#create a plot
fig, ax = plt.subplots()
#plot observation interval
ax.fill_between(pred['Inputs','Light'],
                pred['Observation','Lower'],
                pred['Observation','Upper'],
                alpha=0.3,
                color='C2',
                label='95% Observation Interval')
#plot mean model prediction
ax.plot(pred['Inputs','Light'], pred['Prediction','Mean'], '-',color='C4',label='Mean Observation')
#plot initial dataset
ax.plot(levels0, obs0, 'o', color='C0',label='Initial Data')
ax.plot(levels1, obs1, 'o', color='C1',label='Optimal Data')
ax.set_xlabel('Green Light %')
ax.set_ylabel('Mean Batch GFP Expression (a.u.)')
ax.legend(loc=4)
plt.show()

#######################
# Dynamic model example
#######################

## Supplemental Listing 5 ##############################################
# import casadi as cs
# import nloed as nl
#create state variable vector
states = cs.SX.sym('states',2)
#create control input symbol
inducer = cs.SX.sym('inducer')
#create parameter symbol vector
parameters = cs.SX.sym('parameters',6)
#log-transformed parameters
alpha = cs.exp(parameters[0]) 
K = cs.exp(parameters[1]) 
delta = cs.exp(parameters[2])
beta = cs.exp(parameters[3]) 
L = cs.exp(parameters[4]) 
gamma = cs.exp(parameters[5])
#create symbolic RHS
rhs = cs.vertcat(alpha*inducer/(K + inducer) - delta*states[0],
                 beta*states[0]/(L + states[0]) - gamma*states[1])
#create casadi RHS function
rhs_func = cs.Function('rhs_func',[states,inducer,parameters],[rhs])

## Supplemental Listing 6 ##############################################
#time step size
dt = 1
# Create symbolics for RK4 integration
k1 = rhs_func(states, inducer, parameters)
k2 = rhs_func(states + dt/2.0*k1, inducer, parameters)
k3 = rhs_func(states + dt/2.0*k2, inducer, parameters)
k4 = rhs_func(states + dt*k3, inducer, parameters)
state_step = states + dt/6.0 * (k1 + 2*k2 + 2*k3 + k4)
# Create a function to perform one step of the RK integration
step_func = cs.Function('step_func',[states, inducer, parameters],[state_step])

## Supplemental Listing 7 ##############################################
#create a symbol for the initial inducer level
initial_inducer = cs.SX.sym('init_inducer')
#define the steady state initial states in terms of the initial inducer
init_mrna = (alpha/delta)*initial_inducer/(K+initial_inducer)
ini_prot = (beta/gamma)*init_mrna/(L+init_mrna)
# zip the initial states into a vector
initial_states = cs.vertcat(init_mrna, ini_prot)

## Supplemental Listing 8 ##############################################
#3 samples per cntrl interval, 4 cntrl intervals
#control intervals are 1+2+3=6 steps long:
# cntrl_int1    cntrl_int2    cntrl_int3
#|-1---2-----3||-1---2-----3||-1---2-----3||-1---2-----3|
#set number of control intervals
num_cntrl_intervals = 4
#create a vector for inducer levels in each control interval
inducer_vector = cs.SX.sym('inducer_vec',4)
#define a sample pattern to apply in each control interval
sample_pattern= [1,2,3] 
#lists to store symbols for each sample point, and times of each sample
sample_list, times = [], []
# set the initial states and initialize the step counter
current_state, step_counter = initial_states, 0
#loop over control invervals
for interval in range(num_cntrl_intervals):
  # loop over sample pattern
  for num_stps in sample_pattern:
    #iterate steps indicated by sample pattern
    for k in range(num_stps):
      #propagate the state variables via integration
      current_state = step_func(current_state, inducer_vector[interval], parameters)
      step_counter+=1
    #save the state symbols and times of each sample
    sample_list.append(current_state)
    times.append(step_counter*dt)

## Supplemental Listing 9 ##############################################
#merge all inducer levels into a single inputs vector
inputs = cs.vertcat(initial_inducer,inducer_vector)
# create list for observation structure 
observation_list = []
#create list to store response names
observ_names, observ_type, observ_times = [], [], []
# loop over samples (time points)
for i in range(len(sample_list)):
  #create a unique name for mrna and prot samples
  mrna_name = 'mrna_'+'t'+"{0:0=2d}".format(times[i])
  prot_name = 'prot_'+'t'+"{0:0=2d}".format(times[i])
  #create mean and var tuple for mrna and prot observ.
  mrna_stats = cs.vertcat(sample_list[i][0], 0.005)
  prot_stats = cs.vertcat(sample_list[i][1], 0.005)
  #create casadi function for mrna and prot stats
  mrna_func = cs.Function(mrna_name,[inputs,parameters],[mrna_stats])
  prot_func = cs.Function(prot_name,[inputs,parameters],[prot_stats])
  #append the casadi function and distribution type to obs struct
  observation_list.extend([(mrna_func,'Normal'), (prot_func,'Normal')])
  #store observation names, useful for plotting
  observ_names.extend([mrna_name,prot_name])
  #store observation type
  observ_type.extend(['RNA','Prot'])
  #store observation time
  observ_times.extend([times[i]]*2)

## Supplemental Listing 10 ##############################################
#list the inpit and parameter names
input_names = ['Init_Inducer','Inducer_1','Inducer_2','Inducer_3','Inducer_4']
parameter_names = ['log_Alpha','log_K','log_Delta','log_Beta','log_L','log_Gamma']
#instantiate the model object
model_object = nl.Model(observation_list, input_names, parameter_names)



""" 
Add a docstring
"""
import numpy as np
import pandas as pd
import casadi as cs
from nloed import Model
from nloed import Design

####################################################################################################
# SET UP MODEL
####################################################################################################

#define input and parameters
x = cs.SX.sym('x',2)
xnames = ['Light','Time']
p = cs.SX.sym('p',5)
pnames = ['Basal','MaxInc','Hill','HalfConst','Decay']

#init_param = np.log([2000,8000,2,500,np.log(2)*3/2])
init_param = np.log([442.96,10000,2,995.01,0.27])

#define steady state 
SS = cs.exp(p[0]) + cs.exp(p[1])*x[0]**cs.exp(p[2])/(cs.exp(p[3])**cs.exp(p[2])+x[0]**cs.exp(p[2]))
y_ss = cs.Function('SS',[x,p],[SS])
y0 = 1000 #y_ss(1,init_param)
#define the deterministic model
model = y0 * cs.exp(-cs.exp(p[4])*x[1]) + y_ss(x,p) * (1-cs.exp(-cs.exp(p[4])*x[1]))

#link the deterministic model to the sampling statistics (here normal mean and variance)
mean, var = model, 100**2/3
normal_stats = cs.vertcat(mean, var)

#create a casadi function mapping input and parameters to sampling statistics (mean and var)
y = cs.Function('GFP',[x,p],[normal_stats])

# enter the function as a tuple with label indicating normal error, into observation list
observ_list = [(y,'Normal')]

#instantiate nloed model class
nloed_model = Model(observ_list, xnames, pnames)

####################################################################################################
# Fit
####################################################################################################

obs = [ 1951,	723,	580,
        1479,	1350,	544,
        1516,	680,	543,
        2057,	628,	766,
        1360,	741,	699,
        1670,	640,	570,
        1164,	1135,	2064,
        1419,	1340,	2171,
        1330,	1615,	4598,
        1520,	4024,	8814,
        1657,	3898,	9919,
        1208,	4071,	8860,
        1784,	9213,	9584,
        1168,	5550,	10295,
        1739,	9006,	10672]

init_data = pd.DataFrame({'Light':[.1]*9+[63]*9+[254]*9+[1022]*9+[4095]*9,
                          'Time':[0.5,1.8,6]*15,
                          'Variable':['GFP']*45,
                          'Observation':obs}) 

fit_options={'Confidence':'Intervals',  #NOTE: ISSUE WITH PROFILES HERE
             'InitParamBounds':[(-1,4),(-1,4),(-1,4),(-1,4),(-1,4)],
             'InitSearchNumber':7,
             'SearchBound':5.}
#fit the model to the init data
fit_info = nloed_model.fit(init_data, options=fit_options)

print(np.exp(fit_info))

fit_params = fit_info['Estimate'].to_numpy().flatten()

####################################################################################################
# GENERATE OPTIMAL DESIGN
####################################################################################################

#set up the design algorithm to use continuous (continuous) optimization with two unique inputs points
# continuous_inputs={'Inputs':['Light','Time'],
#                    'Bounds':[(.1,4095.),(1.,6.)],
#                    'Structure':[['L1','T1'],
#                                 ['L2','T2'],
#                                 ['L3','T3'],
#                                 ['L4','T4'],
#                                 ['L5','T5'],
#                                 ['L6','T6'],
#                                 ['L7','T7']],
#                     'Initial':[[10,  5],
#                                [200, 5],
#                                [500, 5],
#                                [4000,5],
#                                [4000,5],
#                                [500, 1],
#                                [4000,1]]}

continuous_inputs={'Inputs':['Light','Time'],
                   'Bounds':[(.1,4095.),(1.,6.)],
                   'Structure':[['L1','T1'],
                                ['L2','T1'],
                                ['L3','T1'],
                                ['L4','T1'],
                                ['L5','T1'],
                                ['L1','T2'],
                                ['L2','T2'],
                                ['L3','T2'],
                                ['L4','T2'],
                                ['L5','T2'],
                                ['L1','T3'],
                                ['L2','T3'],
                                ['L3','T3'],
                                ['L4','T3'],
                                ['L5','T3'],],
                    'Initial':[[.2,  5],
                               [200, 5],
                               [500, 5],
                               [800, 5],
                               [4000,5],
                               [.2,  2],
                               [200, 2],
                               [500, 2],
                               [800, 2],
                               [4000,2],
                               [.2,  1],
                               [200, 1],
                               [500, 1],
                               [800, 1],
                               [4000,1],]}

init_design = pd.DataFrame({'Light':[.1]*9+[63]*9+[254]*9+[1022]*9+[4095]*9,
                            'Time':[0.5,1.8,6]*15,
                            'Variable':['GFP']*45,
                            'Replicats':[1]*45}) 

fixed_design = {'Weight':0.5,'Design':init_design}

opts= {'LockWeights':True}
# generate the optimal discreteimate (relaxed) design
relaxed_design = Design(nloed_model,
                        fit_params,
                        'D',
                        continuous_inputs=continuous_inputs,
                        fixed_design=fixed_design,
                        options=opts)

sample_size = 45
#generate a rounded exact design 
exact_design = relaxed_design.round(sample_size)
print(exact_design)

opt_design_tot = pd.concat([init_design, exact_design], ignore_index=True)

#get estimated covariance, bias and MSE of parameter fit (use asymptotic method here) 
opts={'Method':'MonteCarlo','Covariance':True,'Bias':True,'MSE':True,'SampleNumber':500}
diagnostic_info1 = nloed_model.evaluate(opt_design_tot, fit_params, opts)
print(diagnostic_info1)

naive_design2 = pd.DataFrame({'Light':[102]*3+[161]*3+[256]*3+[406]*3+[644]*3,
                              'Time':[0.3,0.8,3.3]*5,
                              'Variable':['GFP']*15,
                              'Replicats':[1]*15}) 

naive_design_tot = pd.concat([init_design, naive_design2], ignore_index=True)

opts={'Method':'MonteCarlo','Covariance':True,'Bias':True,'MSE':True,'SampleNumber':500}
diagnostic_info2 = nloed_model.evaluate(naive_design_tot, fit_params, opts)
print(diagnostic_info2)

# opts={'Method':'Asymptotic','Covariance':True,'Bias':True,'MSE':True}
# diagnostic_info = nloed_model.evaluate(exact_design, init_param, opts)
# print(diagnostic_info)

####################################################################################################
# GENERATE SAMPLE DATA & FIT MODEL
####################################################################################################

#generate some data, to stand in for an initial experiment
data = nloed_model.sample(exact_design,np.log(init_params))

print(data)

#pass some additional options to fitting alglorithm, including Profile likelihood
fit_options={'Confidence':'Intervals',  #NOTE: ISSUE WITH PROFILES HERE
             'InitParamBounds':[(-1,4),(-1,4),(-1,4),(-1,4)],
             'InitSearchNumber':7,
             'SearchBound':5.}
#fit the model to the init data
fit_info = nloed_model.fit(data, options=fit_options)

print(fit_info)

fit_params = fit_info['Estimate'].to_numpy().flatten()

####################################################################################################
# EVALUATE DESIGN & PREDICT OUTPUT
####################################################################################################



#generate predictions with error bars fdor a random selection of inputs)
prediction_inputs = pd.DataFrame({ 'Light':np.random.uniform(0,5,10),
                                'Variable':['y']*10})
cov_mat = diagnostic_info['Covariance'].to_numpy()
pred_options = {'PredictionInterval':True,
                'ObservationInterval':True,
                'Sensitivity':True}
predictions = nloed_model.predict(prediction_inputs,
                                  fit_params,
                                  covariance_matrix = cov_mat,
                                  options=pred_options)
print(predictions)

t=0



#!/usr/bin/env python
# coding: utf-8

# # Design Evaluations and Model Predictions for the Hill Function Model
# 

# In[1]:


get_ipython().run_cell_magic('capture', '', 'import pandas as pd\nimport matplotlib.pyplot as plt\nfrom N3_sampling_and_fitting import *')


# ### Evaluating the Optimized Design

# In[ ]:





# In[ ]:





# ### Asymptotic Covariance and Confidence Intervals

# In[3]:


eval_options={'Method':'Asymptotic',
              'Covariance':True}
asymptotic_covariance = hill_model.evaluate(exact_design,fit_params,eval_options)
print(asymptotic_covariance)


# In[23]:


asymptotic_lower_bound = fit_params - 2*np.sqrt(np.diag(asymptotic_covariance))
asymptotic_upper_bound = fit_params + 2*np.sqrt(np.diag(asymptotic_covariance))

fim_bnds_array = np.vstack((asymptotic_lower_bound,
                            fit_params,
                            asymptotic_upper_bound)).T
fim_bnds0 = pd.DataFrame(np.exp(fim_bnds_array),
                        index=['Alpha0','Alpha','n','K'],
                        columns=['Lower','Estimate','Upper'])
print('')
print(fim_bnds0)
print('')


# ### Model Predictions

# In[24]:


#convert the covariance matrix to numpy
covariance_matrix = asymptotic_covariance.to_numpy()
#generate predictions with error bars fdor a random selection of inputs)
prediction_inputs = pd.DataFrame({'Inducer':np.linspace(0.1,10,100),
                                  'Variable':['Response']*100})
#request prediction and observation intervals
prediction_options = {'PredictionInterval':True,
                      'ObservationInterval':True}
#generate predictions and intervals
predictions = hill_model.predict(prediction_inputs,
                                   fit_params,
                                   covariance_matrix = covariance_matrix,
                                   options=prediction_options)


# In[ ]:





# ### Plotting Prediction Intervals

# In[25]:


#create plot
fig, ax = plt.subplots()
#plot observation interval
ax.fill_between(predictions['Inputs','Light'],
                predictions['Observation','Lower'],
                predictions['Observation','Upper'],
                alpha=0.3,
                color='C1')
#plot prediction interval
ax.fill_between(predictions['Inputs','Light'],
                predictions['Prediction','Lower'],
                predictions['Prediction','Upper'],
                alpha=0.4,
                color='C0')
#plot mean model prediction
ax.plot(predictions['Inputs','Light'], predictions['Prediction','Mean'], '-')
#plot initial dataset
ax.plot(init_data['Light'], init_data['Observation'], 'o', color='C1')
ax.set_xlabel('Light')
ax.set_ylabel('GFP')
plt.show()


# In[ ]:







#!/usr/bin/env python
# coding: utf-8

# # Optimizing a Design for the Hill Function Model
# 
# 

# In[2]:


#%%capture
import numpy as np
from N1_model_creation import *


# ### Setting the Nominal Parameter Values and Design Objective

# In[3]:


#set nominal parameter values, from previous fit or literature
nominal_param = np.log([1,5,2,1])
#set design objective, here D-optimal design is used
objective = 'D'


# ### Creating a Optimal Design with Continuous Optimized Input Levels

# In[4]:


#create a dictionary describing the design requirments for a continuous design
#here input inducer levels are optimization variables that will be selected
#four unique inducer levels are allowed for in the design optimization
continuous_inputs={'Inputs':['Inducer'],
                   'Bounds':[(.1,5)],
                   'Structure':[['x1'],['x2'],['x3'],['x4']]}


# In[14]:


#instantiate the NLoed design object
continuous_design_object = nl.Design(hill_model, nominal_param, objective, continuous_inputs=continuous_inputs)


# In[6]:


#use the relaxed method to print out the optimal relaxed design
#note, relaxed design's have continuous weights rather than a number of samples for each input-observation pair
continuous_design_object.relaxed()


# In[7]:


#set the sample size for the exact design
sample_size = 48
#use the round function to round the optimal relaxed design to an exact one
exact_continuous_design = continuous_design_object.round(sample_size)
exact_continuous_design


# ### Creating an Optimal Design with Predefined Discrete Input Levels

# In[8]:


#create a dictionary describing the design requirments for a discrete design
#here input inducer levels must be selected from the candidate list
#the design optimization will allocate the number of observations to each input level
discrete_inputs={'Inputs':['Inducer'],
                 'Candidates':[[.1,.25,.5,1,2,3,4,5]]}


# In[9]:


#instantiate the NLoed design object
discrete_design_object = nl.Design(hill_model, nominal_param, objective, discrete_inputs=discrete_inputs)


# In[10]:


#use the relaxed method to print out the optimal relaxed design
discrete_design_object.relaxed()


# In[11]:


#set the sample size for the exact design
sample_size = 48
#use the round function to round the optimal relaxed design to an exact one
exact_discrete_design = discrete_design_object.round(sample_size)
exact_discrete_design




""" 
Add a docstring
"""
import numpy as np
import pandas as pd
import casadi as cs
from nloed import Model
from nloed import Design

####################################################################################################
# SET UP MODEL
####################################################################################################

#define input and parameters
x = cs.SX.sym('x',1)
xnames = ['Light']
p = cs.SX.sym('p',4)
pnames = ['Basal','Rate','Hill','HalfConst']

#define the deterministic model
hull_func = cs.exp(p[0]) + cs.exp(p[1])*x[0]**cs.exp(p[2])/(cs.exp(p[3])**cs.exp(p[2])+x[0]**cs.exp(p[2]))

#link the deterministic model to the sampling statistics (here normal mean and variance)
mean, var = hull_func, 0.1
normal_stats = cs.vertcat(mean, var)

#create a casadi function mapping input and parameters to sampling statistics (mean and var)
y = cs.Function('y',[x,p],[normal_stats])

# enter the function as a tuple with label indicating normal error, into observation list
observ_list = [(y,'Normal')]

#instantiate nloed model class
nloed_model = Model(observ_list,xnames,pnames)

####################################################################################################
# GENERATE OPTIMAL DESIGN
####################################################################################################

#set up the design algorithm to use continuous (continuous) optimization with two unique inputs points
continuous_inputs={'Inputs':['Light'],'Bounds':[(.1,5)],'Structure':[['x1'],['x2'],['x3'],['x4']]}

true_param = [1,5,2,1]
# generate the optimal discreteimate (relaxed) design
relaxed_design = Design(nloed_model,np.log(true_param),'D',continuous_inputs=continuous_inputs)

sample_size = 30
#generate a rounded exact design 
exact_design = relaxed_design.round(sample_size)

print(exact_design)

####################################################################################################
# GENERATE SAMPLE DATA & FIT MODEL
####################################################################################################

#generate some data, to stand in for an initial experiment
data = nloed_model.sample(exact_design,np.log(true_param))

print(data)

#pass some additional options to fitting alglorithm, including Profile likelihood
fit_options={'Confidence':'Intervals',  #NOTE: ISSUE WITH PROFILES HERE
             'InitParamBounds':[(-1,4),(-1,4),(-1,4),(-1,4)],
             'InitSearchNumber':7,
             'SearchBound':5.}
#fit the model to the init data
fit_info = nloed_model.fit(data, options=fit_options)

print(fit_info)

fit_params = fit_info['Estimate'].to_numpy().flatten()

####################################################################################################
# EVALUATE DESIGN & PREDICT OUTPUT
####################################################################################################

#get estimated covariance, bias and MSE of parameter fit (use asymptotic method here) 
opts={'Covariance':True,'Bias':True,'MSE':True,'SampleNumber':100}
diagnostic_info = nloed_model.evaluate(exact_design,fit_params,opts)
print(diagnostic_info)

#generate predictions with error bars fdor a random selection of inputs)
prediction_inputs = pd.DataFrame({ 'Light':np.random.uniform(0,5,10),
                                'Variable':['y']*10})
cov_mat = diagnostic_info['Covariance'].to_numpy()
pred_options = {'PredictionInterval':True,
                'ObservationInterval':True,
                'Sensitivity':True}
predictions = nloed_model.predict(prediction_inputs,
                                  fit_params,
                                  covariance_matrix = cov_mat,
                                  options=pred_options)
print(predictions)

t=0



#!/usr/bin/env python
# coding: utf-8

# # Creating a Simple Hill Function Model
# 
# 

# In[8]:


#import nloed and casadi
import nloed as nl
import casadi as cs


# ### Model Parameterization and Equations

# In[9]:


#create symbol for the model input
x = cs.SX.sym('x',1)
#create symbols for the model parameters
p = cs.SX.sym('p',4)


# In[10]:


#use parameter transformation to ensure named parameter remain positive
#we will fit and design for fitting p, which is exponent of the 'natural' parameters
alpha_0 = cs.exp(p[0])
alpha = cs.exp(p[1])
K = cs.exp(p[2])
n = cs.exp(p[3])


# In[11]:


#write a symbolic expression for the model, predicting mean observations from input and parameters
mean = alpha_0 + alpha*x[0]**n / (K**n + x[0]**n)


# In[12]:


#create an symbolic expression for the variance of the observations
#here variance is assumed to be proportional to the mean i.e. hetroskedastic
var = 0.01*mean


# ### Preparing to Call the NLoed Model Constructor

# In[13]:


#group the mean and variance into a single vecotor
normal_stats = cs.vertcat(mean, var)
#create a CasADi function to predict the mean and variance of the observation from the input and parameters
model_function = cs.Function('Observation',[x,p],[normal_stats])


# In[14]:


#list user-chosen names of model inputs (only one in this example)
xnames = ['Inducer']
#list user-chosen names of parameters, log_ prefix is to remind us of transformation
pnames = ['log_Basal','log_Rate','log_Hill','log_HalfConst']
#create a list of model observation variable (only one)
#the single list element is a tuple of the model function and a lable indicating the distribution type
observ_list = [(model_function,'Normal')]
#create the NLoed model using the Model constructor
hill_model = nl.Model(observ_list,xnames,pnames)

